@@ -20,6 +20,7 @@
  */
 
 #include "terminal-intl.h"
+#include "terminal.h"
 
 #include "encoding.h"
 
@@ -52,164 +53,163 @@ static GConfClient *default_client = NULL;
 
 static TerminalEncoding encodings[] = {
 
-  { TERMINAL_ENCODING_CURRENT_LOCALE,
+  { TERMINAL_ENCODING_CURRENT_LOCALE, TRUE,
     NULL, N_("Current Locale") },
 
-  { TERMINAL_ENCODING_ISO_8859_1,
+  { TERMINAL_ENCODING_ISO_8859_1, FALSE,
     "ISO-8859-1", N_("Western") },
-  { TERMINAL_ENCODING_ISO_8859_3,
+  { TERMINAL_ENCODING_ISO_8859_2, FALSE,
+    "ISO-8859-2", N_("Central European") },
+  { TERMINAL_ENCODING_ISO_8859_3, FALSE,
     "ISO-8859-3", N_("South European") },
-  { TERMINAL_ENCODING_ISO_8859_4,
+  { TERMINAL_ENCODING_ISO_8859_4, FALSE,
     "ISO-8859-4", N_("Baltic") },
-  { TERMINAL_ENCODING_ISO_8859_5,
+  { TERMINAL_ENCODING_ISO_8859_5, FALSE,
     "ISO-8859-5", N_("Cyrillic") },
-  { TERMINAL_ENCODING_ISO_8859_6,
+  { TERMINAL_ENCODING_ISO_8859_6, FALSE,
     "ISO-8859-6", N_("Arabic") },
-  { TERMINAL_ENCODING_ISO_8859_7,
+  { TERMINAL_ENCODING_ISO_8859_7, FALSE,
     "ISO-8859-7", N_("Greek") },
-  { TERMINAL_ENCODING_ISO_8859_8,
+  { TERMINAL_ENCODING_ISO_8859_8, FALSE,
     "ISO-8859-8", N_("Hebrew Visual") },
-  { TERMINAL_ENCODING_ISO_8859_8_I,
+  { TERMINAL_ENCODING_ISO_8859_8_I, FALSE,
     "ISO-8859-8-I", N_("Hebrew") },
-  { TERMINAL_ENCODING_ISO_8859_9,
+  { TERMINAL_ENCODING_ISO_8859_9, FALSE,
     "ISO-8859-9", N_("Turkish") },
-  { TERMINAL_ENCODING_ISO_8859_10,
+  { TERMINAL_ENCODING_ISO_8859_10, FALSE,
     "ISO-8859-10", N_("Nordic") },
-  { TERMINAL_ENCODING_ISO_8859_13,
+  { TERMINAL_ENCODING_ISO_8859_13, FALSE,
     "ISO-8859-13", N_("Baltic") },
-  { TERMINAL_ENCODING_ISO_8859_14,
+  { TERMINAL_ENCODING_ISO_8859_14, FALSE,
     "ISO-8859-14", N_("Celtic") },
-  { TERMINAL_ENCODING_ISO_8859_15,
+  { TERMINAL_ENCODING_ISO_8859_15, FALSE,
     "ISO-8859-15", N_("Western") },
-  { TERMINAL_ENCODING_ISO_8859_16,
+  { TERMINAL_ENCODING_ISO_8859_16, FALSE,
     "ISO-8859-16", N_("Romanian") },
 
-  { TERMINAL_ENCODING_UTF_7,
+  { TERMINAL_ENCODING_UTF_7, FALSE,
     "UTF-7", N_("Unicode") },
-  { TERMINAL_ENCODING_UTF_8,
+  { TERMINAL_ENCODING_UTF_8, FALSE,
     "UTF-8", N_("Unicode") },
-  { TERMINAL_ENCODING_UTF_16,
+  { TERMINAL_ENCODING_UTF_16, FALSE,
     "UTF-16", N_("Unicode") },
-  { TERMINAL_ENCODING_UCS_2,
+  { TERMINAL_ENCODING_UCS_2, FALSE,
     "UCS-2", N_("Unicode") },
-  { TERMINAL_ENCODING_UCS_4,
+  { TERMINAL_ENCODING_UCS_4, FALSE,
     "UCS-4", N_("Unicode") },
 
-  { TERMINAL_ENCODING_ARMSCII_8,
+  { TERMINAL_ENCODING_ARMSCII_8, FALSE,
     "ARMSCII-8", N_("Armenian") },
-  { TERMINAL_ENCODING_BIG5,
+  { TERMINAL_ENCODING_BIG5, FALSE,
     "BIG5", N_("Chinese Traditional") },
-  { TERMINAL_ENCODING_BIG5_HKSCS,
+  { TERMINAL_ENCODING_BIG5_HKSCS, FALSE,
     "BIG5-HKSCS", N_("Chinese Traditional") },
-  { TERMINAL_ENCODING_CP_866,
+  { TERMINAL_ENCODING_CP_866, FALSE,
     "CP866", N_("Cyrillic/Russian") },
 
-  { TERMINAL_ENCODING_EUC_JP,
+  { TERMINAL_ENCODING_EUC_JP, FALSE,
     "EUC-JP", N_("Japanese") },
-  { TERMINAL_ENCODING_EUC_KR,
+  { TERMINAL_ENCODING_EUC_KR, FALSE,
     "EUC-KR", N_("Korean") },
-  { TERMINAL_ENCODING_EUC_TW,
+  { TERMINAL_ENCODING_EUC_TW, FALSE,
     "EUC-TW", N_("Chinese Traditional") },
 
-  { TERMINAL_ENCODING_GB18030,
+  { TERMINAL_ENCODING_GB18030, FALSE,
     "GB18030", N_("Chinese Simplified") },
-  { TERMINAL_ENCODING_GB2312,
+  { TERMINAL_ENCODING_GB2312, FALSE,
     "GB2312", N_("Chinese Simplified") },
-  { TERMINAL_ENCODING_GBK,
+  { TERMINAL_ENCODING_GBK, FALSE,
     "GBK", N_("Chinese Simplified") },
-  { TERMINAL_ENCODING_GEOSTD8,
+  { TERMINAL_ENCODING_GEOSTD8, FALSE,
     "GEORGIAN-ACADEMY", N_("Georgian") }, /* FIXME GEOSTD8 ? */
-  { TERMINAL_ENCODING_HZ,
+  { TERMINAL_ENCODING_HZ, FALSE,
     "HZ", N_("Chinese Simplified") },
 
-  { TERMINAL_ENCODING_IBM_850,
+  { TERMINAL_ENCODING_IBM_850, FALSE,
     "IBM850", N_("Western") },
-  { TERMINAL_ENCODING_IBM_852,
+  { TERMINAL_ENCODING_IBM_852, FALSE,
     "IBM852", N_("Central European") },
-  { TERMINAL_ENCODING_IBM_855,
+  { TERMINAL_ENCODING_IBM_855, FALSE,
     "IBM855", N_("Cyrillic") },
-  { TERMINAL_ENCODING_IBM_857,
+  { TERMINAL_ENCODING_IBM_857, FALSE,
     "IBM857", N_("Turkish") },
-  { TERMINAL_ENCODING_IBM_862,
+  { TERMINAL_ENCODING_IBM_862, FALSE,
     "IBM862", N_("Hebrew") },
-  { TERMINAL_ENCODING_IBM_864,
+  { TERMINAL_ENCODING_IBM_864, FALSE,
     "IBM864", N_("Arabic") },
 
-  { TERMINAL_ENCODING_ISO_2022_JP,
+  { TERMINAL_ENCODING_ISO_2022_JP, FALSE,
     "ISO2022JP", N_("Japanese") },
-  { TERMINAL_ENCODING_ISO_2022_KR,
+  { TERMINAL_ENCODING_ISO_2022_KR, FALSE,
     "ISO2022KR", N_("Korean") },
-  { TERMINAL_ENCODING_ISO_IR_111,
+  { TERMINAL_ENCODING_ISO_IR_111, FALSE,
     "ISO-IR-111", N_("Cyrillic") },
-  { TERMINAL_ENCODING_JOHAB,
+  { TERMINAL_ENCODING_JOHAB, FALSE,
     "JOHAB", N_("Korean") },
-  { TERMINAL_ENCODING_KOI8_R,
+  { TERMINAL_ENCODING_KOI8_R, FALSE,
     "KOI8R", N_("Cyrillic") },
-  { TERMINAL_ENCODING_KOI8_U,
+  { TERMINAL_ENCODING_KOI8_U, FALSE,
     "KOI8U", N_("Cyrillic/Ukrainian") },
 
-#if 0
-  /* GLIBC iconv doesn't seem to have these mac things */
-  { TERMINAL_ENCODING_MAC_ARABIC,
+  { TERMINAL_ENCODING_MAC_ARABIC, FALSE,
     "MAC_ARABIC", N_("Arabic") },
-  { TERMINAL_ENCODING_MAC_CE,
+  { TERMINAL_ENCODING_MAC_CE, FALSE,
     "MAC_CE", N_("Central European") },
-  { TERMINAL_ENCODING_MAC_CROATIAN,
+  { TERMINAL_ENCODING_MAC_CROATIAN, FALSE,
     "MAC_CROATIAN", N_("Croatian") },
-  { TERMINAL_ENCODING_MAC_CYRILLIC,
+  { TERMINAL_ENCODING_MAC_CYRILLIC, FALSE,
     "MAC-CYRILLIC", N_("Cyrillic") },
-  { TERMINAL_ENCODING_MAC_DEVANAGARI,
+  { TERMINAL_ENCODING_MAC_DEVANAGARI, FALSE,
     "MAC_DEVANAGARI", N_("Hindi") },
-  { TERMINAL_ENCODING_MAC_FARSI,
+  { TERMINAL_ENCODING_MAC_FARSI, FALSE,
     "MAC_FARSI", N_("Farsi") },
-  { TERMINAL_ENCODING_MAC_GREEK,
+  { TERMINAL_ENCODING_MAC_GREEK, FALSE,
     "MAC_GREEK", N_("Greek") },
-  { TERMINAL_ENCODING_MAC_GUJARATI,
+  { TERMINAL_ENCODING_MAC_GUJARATI, FALSE,
     "MAC_GUJARATI", N_("Gujarati") },
-  { TERMINAL_ENCODING_MAC_GURMUKHI,
+  { TERMINAL_ENCODING_MAC_GURMUKHI, FALSE,
     "MAC_GURMUKHI", N_("Gurmukhi") },
-  { TERMINAL_ENCODING_MAC_HEBREW,
+  { TERMINAL_ENCODING_MAC_HEBREW, FALSE,
     "MAC_HEBREW", N_("Hebrew") },
-  { TERMINAL_ENCODING_MAC_ICELANDIC,
+  { TERMINAL_ENCODING_MAC_ICELANDIC, FALSE,
     "MAC_ICELANDIC", N_("Icelandic") },
-  { TERMINAL_ENCODING_MAC_ROMAN,
+  { TERMINAL_ENCODING_MAC_ROMAN, FALSE,
     "MAC_ROMAN", N_("Western") },
-  { TERMINAL_ENCODING_MAC_ROMANIAN,
+  { TERMINAL_ENCODING_MAC_ROMANIAN, FALSE,
     "MAC_ROMANIAN", N_("Romanian") },
-  { TERMINAL_ENCODING_MAC_TURKISH,
+  { TERMINAL_ENCODING_MAC_TURKISH, FALSE,
     "MAC_TURKISH", N_("Turkish") },
-  { TERMINAL_ENCODING_MAC_UKRAINIAN,
+  { TERMINAL_ENCODING_MAC_UKRAINIAN, FALSE,
     "MAC_UKRAINIAN", N_("Cyrillic/Ukrainian") },
-#endif
   
-  { TERMINAL_ENCODING_SHIFT_JIS,
+  { TERMINAL_ENCODING_SHIFT_JIS, FALSE,
     "SHIFT-JIS", N_("Japanese") },
-  { TERMINAL_ENCODING_TCVN,
+  { TERMINAL_ENCODING_TCVN, FALSE,
     "TCVN", N_("Vietnamese") },
-  { TERMINAL_ENCODING_TIS_620,
+  { TERMINAL_ENCODING_TIS_620, FALSE,
     "TIS-620", N_("Thai") },
-  { TERMINAL_ENCODING_UHC,
+  { TERMINAL_ENCODING_UHC, FALSE,
     "UHC", N_("Korean") },
-  { TERMINAL_ENCODING_VISCII,
+  { TERMINAL_ENCODING_VISCII, FALSE,
     "VISCII", N_("Vietnamese") },
 
-  { TERMINAL_ENCODING_WINDOWS_1250,
+  { TERMINAL_ENCODING_WINDOWS_1250, FALSE,
     "WINDOWS-1250", N_("Central European") },
-  { TERMINAL_ENCODING_WINDOWS_1251,
+  { TERMINAL_ENCODING_WINDOWS_1251, FALSE,
     "WINDOWS-1251", N_("Cyrillic") },
-  { TERMINAL_ENCODING_WINDOWS_1252,
+  { TERMINAL_ENCODING_WINDOWS_1252, FALSE,
     "WINDOWS-1252", N_("Western") },
-  { TERMINAL_ENCODING_WINDOWS_1253,
+  { TERMINAL_ENCODING_WINDOWS_1253, FALSE,
     "WINDOWS-1253", N_("Greek") },
-  { TERMINAL_ENCODING_WINDOWS_1254,
+  { TERMINAL_ENCODING_WINDOWS_1254, FALSE,
     "WINDOWS-1254", N_("Turkish") },
-  { TERMINAL_ENCODING_WINDOWS_1255,
+  { TERMINAL_ENCODING_WINDOWS_1255, FALSE,
     "WINDOWS-1255", N_("Hebrew") },
-  { TERMINAL_ENCODING_WINDOWS_1256,
+  { TERMINAL_ENCODING_WINDOWS_1256, FALSE,
     "WINDOWS-1256", N_("Arabic") },
-  { TERMINAL_ENCODING_WINDOWS_1257,
+  { TERMINAL_ENCODING_WINDOWS_1257, FALSE,
     "WINDOWS-1257", N_("Baltic") },
-  { TERMINAL_ENCODING_WINDOWS_1258,
+  { TERMINAL_ENCODING_WINDOWS_1258, FALSE,
     "WINDOWS-1258", N_("Vietnamese") }
 };
 
@@ -304,6 +304,7 @@ terminal_encoding_copy (const TerminalEncoding *src)
 
   c = g_new (TerminalEncoding, 1);
   c->index = src->index;
+  c->valid = src->valid;
   c->name = g_strdup (src->name);
   c->charset = g_strdup (src->charset);
   
@@ -314,6 +315,8 @@ static void
 update_active_encodings_from_string_list (GSList *strings)
 {
   GSList *tmp;
+  GHashTable *table;
+  const char *charset;
 
 #if 1
   g_slist_foreach (active_encodings, (GFunc) terminal_encoding_free,
@@ -321,50 +324,75 @@ update_active_encodings_from_string_list (GSList *strings)
   g_slist_free (active_encodings);
 #endif
   active_encodings = NULL;
+
+  table = g_hash_table_new (g_direct_hash, g_direct_equal);
   
-  tmp = strings;
-  while (tmp != NULL)
+  /* First add the local encoding. */
+  charset = encodings[TERMINAL_ENCODING_CURRENT_LOCALE].charset;
+  if (g_hash_table_lookup (table, GINT_TO_POINTER (g_quark_from_string (charset))) == NULL)
+    {
+      active_encodings = g_slist_prepend (active_encodings,
+                                          terminal_encoding_copy (&encodings[TERMINAL_ENCODING_CURRENT_LOCALE]));
+      g_hash_table_insert (table,
+		           GINT_TO_POINTER (g_quark_from_string (charset)),
+		           GINT_TO_POINTER (g_quark_from_string (charset)));
+    }
+
+  /* Always ensure that UTF-8 is available. */
+  charset = encodings[TERMINAL_ENCODING_UTF_8].charset;
+  if (g_hash_table_lookup (table, GINT_TO_POINTER (g_quark_from_string (charset))) == NULL)
+    {
+      active_encodings = g_slist_prepend (active_encodings,
+                                          terminal_encoding_copy (&encodings[TERMINAL_ENCODING_UTF_8]));
+      g_hash_table_insert (table,
+		           GINT_TO_POINTER (g_quark_from_string (charset)),
+		           GINT_TO_POINTER (g_quark_from_string (charset)));
+    }
+
+  for (tmp = strings; tmp != NULL; tmp = tmp->next)
     {
       const TerminalEncoding *e;
-      const char *charset = tmp->data;
+      charset = tmp->data;
       TerminalEncoding *encoding;
       
       if (strcmp (charset, "current") == 0)
         g_get_charset (&charset);
       
       e = find_encoding_by_charset (charset);
+
+      if (g_hash_table_lookup (table, GINT_TO_POINTER (g_quark_from_string (charset))) != NULL)
+        {
+	  continue;
+        }
+
+      g_hash_table_insert (table,
+		           GINT_TO_POINTER (g_quark_from_string (charset)),
+		           GINT_TO_POINTER (g_quark_from_string (charset)));
       
       if (e == NULL)
         {
           encoding = g_new0 (TerminalEncoding, 1);
           
           encoding->index = -1;
+          encoding->valid = TRUE; /* scary! */
           encoding->charset = g_strdup (charset);
           encoding->name = g_strdup (_("User Defined"));
         }
       else
         {
-          encoding = terminal_encoding_copy (e);
+          encoding = e->valid ? terminal_encoding_copy (e) : NULL;
         }
 
-      active_encodings = g_slist_prepend (active_encodings, encoding);
-      
-      tmp = tmp->next;
+      if (encoding != NULL)
+        {
+          active_encodings = g_slist_prepend (active_encodings, encoding);
+        }
     }
 
   /* Put it back in order, order is significant */
   active_encodings = g_slist_reverse (active_encodings);
   
-  if (active_encodings == NULL)
-    {
-      /* Emergency fallbacks */
-      active_encodings = g_slist_prepend (active_encodings,
-                                          terminal_encoding_copy (&encodings[TERMINAL_ENCODING_CURRENT_LOCALE]));
-      if (strcmp (encodings[TERMINAL_ENCODING_CURRENT_LOCALE].charset,
-                  "UTF-8") != 0)
-        active_encodings = g_slist_prepend (active_encodings,
-                                            terminal_encoding_copy (&encodings[TERMINAL_ENCODING_UTF_8]));
-    }
+  g_hash_table_destroy (table);
   
   update_active_encoding_tree_models ();
 }
@@ -689,49 +717,12 @@ terminal_encoding_dialog_new (GtkWindow *transient_parent)
   GtkTreeIter parent_iter;
   GtkTreeSelection *selection;
   GtkWidget *dialog;
-  
-  if (g_file_test ("./"TERM_GLADE_FILE,
-                   G_FILE_TEST_EXISTS))
-    {
-      /* Try current dir, for debugging */
-      xml = glade_xml_new ("./"TERM_GLADE_FILE,
-                           "encodings-dialog",
-                           GETTEXT_PACKAGE);
-    }
-  else
-    {
-      xml = glade_xml_new (TERM_GLADE_DIR"/"TERM_GLADE_FILE,
-                           "encodings-dialog",
-                           GETTEXT_PACKAGE);
-    }
 
+  xml = terminal_util_load_glade_file (TERM_GLADE_FILE,
+                                       "encodings-dialog",
+                                       transient_parent);
   if (xml == NULL)
-    {
-      static GtkWidget *no_glade_dialog = NULL;
-          
-      if (no_glade_dialog == NULL)
-        {
-          no_glade_dialog =
-            gtk_message_dialog_new (transient_parent,
-                                    GTK_DIALOG_DESTROY_WITH_PARENT,
-                                    GTK_MESSAGE_ERROR,
-                                    GTK_BUTTONS_CLOSE,
-                                    _("The file \"%s\" is missing. This indicates that the application is installed incorrectly, so the encodings dialog can't be displayed."),
-                                    TERM_GLADE_DIR"/"TERM_GLADE_FILE);
-                                        
-          g_signal_connect (G_OBJECT (no_glade_dialog),
-                            "response",
-                            G_CALLBACK (gtk_widget_destroy),
-                            NULL);
-          
-          g_object_add_weak_pointer (G_OBJECT (no_glade_dialog),
-                                     (void**)&no_glade_dialog);
-        }
-
-      gtk_window_present (GTK_WINDOW (no_glade_dialog));
-
-      return NULL;
-    }  
+    return NULL;
 
   /* The dialog itself */
   dialog = glade_xml_get_widget (xml, "encodings-dialog");
@@ -791,14 +782,16 @@ terminal_encoding_dialog_new (GtkWindow *transient_parent)
   i = 0;
   while (i < (int) G_N_ELEMENTS (encodings))
     {
-      gtk_list_store_append (tree, &parent_iter);
-      gtk_list_store_set (tree, &parent_iter,
-                          COLUMN_CHARSET,
-                          encodings[i].charset,
-                          COLUMN_NAME,
-                          encodings[i].name,
-                          -1);
-
+      if (encodings[i].valid)
+        {
+          gtk_list_store_append (tree, &parent_iter);
+          gtk_list_store_set (tree, &parent_iter,
+                              COLUMN_CHARSET,
+                              encodings[i].charset,
+                              COLUMN_NAME,
+                              encodings[i].name,
+                              -1);
+        }
       ++i;
     }
 
@@ -869,6 +862,8 @@ terminal_encoding_dialog_new (GtkWindow *transient_parent)
   g_signal_connect (G_OBJECT (selection), "changed",
                     G_CALLBACK (displayed_selection_changed_callback),
                     dialog);
+
+  g_object_unref (G_OBJECT (xml));
   
   return dialog;
 }
@@ -933,6 +928,9 @@ terminal_encoding_init (GConfClient *conf)
   int i;
   GError *err;
   GSList *strings;
+  gsize bytes_read, bytes_written;
+  gchar *converted;
+  gchar ascii_sample[96];
   
   g_return_if_fail (GCONF_IS_CLIENT (conf));
 
@@ -943,15 +941,56 @@ terminal_encoding_init (GConfClient *conf)
                  &encodings[TERMINAL_ENCODING_CURRENT_LOCALE].charset);
 
   g_assert (G_N_ELEMENTS (encodings) == TERMINAL_ENCODING_LAST);
+
+  /* Initialize the sample text with all of the printing ASCII characters
+   * from space (32) to the tilde (126), 95 in all. */ 
+  for (i = 0; i < sizeof (ascii_sample); i++) 
+    ascii_sample[i] = i + 32;
+
+  ascii_sample[sizeof(ascii_sample) - 1] = '\0';
   
   i = 0;
   while (i < TERMINAL_ENCODING_LAST)
     {
+      bytes_read = 0;
+      bytes_written = 0;
+      
       g_assert (encodings[i].index == i);
 
       /* Translate the names */
       encodings[i].name = _(encodings[i].name);
+
+      /* Test that the encoding is a proper superset of ASCII (which naive
+       * apps are going to use anyway) by attempting to validate the text
+       * using the current encoding.  This also flushes out any encodings
+       * which the underlying GIConv implementation can't support.
+       */
+      converted = g_convert (ascii_sample, sizeof (ascii_sample) - 1,
+		             encodings[i].charset, encodings[i].charset,
+			     &bytes_read, &bytes_written, NULL);
       
+      /* The encoding is only valid if ASCII passes through cleanly. */
+      if (i == TERMINAL_ENCODING_CURRENT_LOCALE)
+        encodings[i].valid = TRUE;
+      else
+        encodings[i].valid =
+          (bytes_read == (sizeof (ascii_sample) - 1)) &&
+          (converted != NULL) &&
+          (strcmp (converted, ascii_sample) == 0);
+
+#ifdef DEBUG_ENCODINGS
+      if (!encodings[i].valid)
+        {
+          g_print("Rejecting encoding %s as invalid:\n", encodings[i].charset);
+          g_print(" input  \"%s\"\n", ascii_sample);
+          g_print(" output \"%s\"\n\n", converted ? converted : "(null)");
+        }
+#endif
+
+      /* Discard the converted string. */
+      if (converted != NULL)
+        g_free (converted);
+
       ++i;
     }
 
